﻿Public Class mAlphaBeta
    Public pos As mPosition                         '评价
    Private Const MAX_GEN_MOVES As Integer = 224    '最大的生成走法步数
    Private Const MATE_VALUE = 10000                '最高分值，即将死的分值
    Private Const LIMIT_DEPTH_SearchFull = 16       '最大的搜索深度
    Private Const WIN_VALUE = 5000                  '搜索出胜负的分值界限，超出此值就说明已经搜索出杀棋了
    Public winplayr As Integer = 2                  '获胜玩家
    Private mCompare As New mvsCompare              '排序器
    Private Const NULL_DEPTH = 2                    '空步裁剪的裁剪深度

    '用局面类初始化
    Sub New(p As mPosition)
        pos = p
    End Sub

    '超出边界(Fail-Soft)的Alpha-Beta搜索过程
    Public Function SearchFull(vlAlpha As Integer, vlBeta As Integer, nDepth As Integer) As Integer
        Dim i, j As Integer
        Dim nGenMoves As Integer = -1
        Dim vl, vlBest, mvBest As Integer
        Dim mvs(MAX_GEN_MOVES) As Byte
        '一个Alpha-Beta完全搜索分为以下几个阶段

        '====================================空步剪裁=====================================
        If pos.nDistance > 0 Then
            '1. 到达水平线
            If nDepth <= 0 Then Return pos.Evaluate '
            '1-1. 到达极限深度就返回局面评价
            If pos.nDistance = LIMIT_DEPTH_SearchFull Then Return pos.Evaluate()
            '1-2. 尝试空步裁剪(根节点的Beta值是"MATE_VALUE"，所以不可能发生空步裁剪)
            If pos.Evaluate() = -3000 Then          '被冲棋时，根据冲棋点返回值生成走法，而不是生成全部走法。
                '遍历棋型信息，提取全部冲棋点。
                For i = 0 To pos.Vectors.lnkinf.Count - 1
                    For j = 0 To pos.Vectors.lnkinf(i).cqpend
                        nGenMoves += 1
                        mvs(nGenMoves) = pos.Vectors.lnkinf(i).cqp(j)
                    Next
                Next
            Else                                    '未被冲棋时进行空步剪裁
                pos.NullMove()
                vl = -SearchFull(-vlBeta, 1 - vlBeta, nDepth - NULL_DEPTH - 1)
                pos.UnNullMove()
                If (vl >= vlBeta) Then
                    Return vl
                End If
            End If
        End If
        '==================================空步剪裁结束===================================
        '2. 初始化最佳值和最佳走法
        vlBest = -MATE_VALUE '这样可以知道，是否一个走法都没走过(杀棋)
        mvBest = 0 '这样可以知道，是否搜索到了Beta走法或PV走法，以便保存到历史表

        '3. 生成全部走法，并根据历史表排序
        '====================检测nGenMoves，用以判断是否在空步剪裁测试时检测到被冲棋并返回了走法=================
        If nGenMoves = -1 Then nGenMoves = pos.GenerateMoves(mvs)
        '==================检测nGenMoves，用以判断是否在空步剪裁测试时检测到被冲棋并返回了走法结束=================
        Array.Sort(mvs, 0, nGenMoves, mCompare)

        '4. 逐一走这些走法，并进行递归
        For i = 0 To nGenMoves
            Dim player = pos.sdPlayer
            pos.AddPiece(mvs(i))
            '=================================冲棋延伸================================
            'vl = -SearchFull(-vlBeta, -vlAlpha, nDepth - 1)
            '=========以上为被替换代码==========
            vl = -SearchFull(-vlBeta, -vlAlpha, IIf(pos.Evaluate = -3000, nDepth, nDepth - 1))
            '===============================冲棋延伸结束==============================
            pos.DelPiece(mvs(i))

            ' 5. 进行Alpha-Beta大小判断和截断
            If (vl > vlBest) Then       '找到最佳值(但不能确定是Alpha、PV还是Beta走法)
                vlBest = vl             '"vlBest"就是目前要返回的最佳值，可能超出Alpha-Beta边界
                If (vl >= vlBeta) Then  '找到一个Beta走法
                    mvBest = mvs(i)     'Beta走法要保存到历史表
                    Exit For            'Beta截断
                End If
                If (vl > vlAlpha) Then  '找到一个PV走法
                    mvBest = mvs(i)     'PV走法要保存到历史表
                    vlAlpha = vl        '缩小Alpha-Beta边界
                End If
            End If
        Next

        '5. 所有走法都搜索完了，把最佳走法(不能是Alpha走法)保存到历史表，返回最佳值
        If vlBest = -MATE_VALUE Then
            '如果是杀棋，就根据杀棋步数给出评价
            Return pos.nDistance - MATE_VALUE
        End If
        If mvBest <> 0 Then
            '如果不是Alpha走法，就将最佳走法保存到历史表
            pos.nHistoryTable(mvBest) += nDepth * nDepth
            If pos.nDistance = 0 Then
                '搜索根节点时，总是有一个最佳走法(因为全窗口搜索不会超出边界)，将这个走法保存下来
                pos.mvResult = mvBest
            End If
        End If
        Return vlBest
    End Function

    '===============================迭代加深===============================
    '迭代加深搜索过程
    Function SearchMain() As Integer
        Dim i, t, vl As Integer
        '初始化
        pos.ClearnHistoryTable() ' 清空历史表
        mvsCompare.ms = pos.nHistoryTable
        t = My.Computer.Clock.TickCount ' 初始化定时器
        pos.nDistance = 0   '初始步数
        winplayr = 2        '胜利者
        '迭代加深过程
        For i = 1 To LIMIT_DEPTH_SearchFull - 1
            Debug.Print("正在迭代：" & i)
            vl = SearchFull(-MATE_VALUE, MATE_VALUE, i)
            '搜索到杀棋，就终止搜索
            If vl > WIN_VALUE Then  '计算机胜利
                winplayr = 1
                Exit For
            End If
            If vl < -WIN_VALUE Then '玩家胜利
                winplayr = 0
                Exit For
            End If
            '超过一秒，就终止搜索
            If My.Computer.Clock.TickCount - t > 1000 Then
                Exit For
            End If
        Next
        Debug.Print("迭代加深:" & i)
        Return pos.mvResult
    End Function
    '==============================================================================

End Class

'根据历史走法对合理招法进行排序的排序器。
Class mvsCompare
    Implements IComparer
    Public Shared ms() As Integer
    Public Function Compare(x As Object, y As Object) As Integer Implements System.Collections.IComparer.Compare
        Return ms(y) - ms(x)
    End Function
End Class